@@ -25,6 +25,10 @@
 #ifndef NGHTTP2_PQ_TEST_H
 #define NGHTTP2_PQ_TEST_H
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif /* HAVE_CONFIG_H */
+
 void test_nghttp2_pq(void);
 void test_nghttp2_pq_update(void);
 void test_nghttp2_pq_remove(void);
